# -*- coding: iso-8859-1 -*-
"""
A little GUI for exploring the power function of simple tests based on
the binomial distribution (e.g., the tea testing lady).
"""

import operator
import math
import bisect
import Tkinter

def faculty(n):
	return reduce(operator.mul, range(1, n+1), 1)

def binomial(up, down):
	return faculty(up)/(faculty(down)*faculty(up-down))

def binomialDist(p, n, N):
	n = int(n)
	return binomial(N, n)*p**n*(1-p)**(N-n)


def computeTicPositions(valRange, maxTics=11):
	"""returns a list of "suitable" tics for valRange.

	Tics are suitable if their values are expressible with fairly
	few decimal digits.  maxTics should be between 5 and 50 or so, otherwise
	results may not be what you expect.

	Here's our basic algorithm: Determine the dimension of tic
	size by int'ing the log of 1/10 of range we are to display.
	This will give between 10 and 100 tics.  We then progressively
	increase the tic distance until we have less than maxTics tics.

	The base tic is the first tic before valRange[0] that is an integer
	multiple of the tic distance.
	>>> computeTicPositions([0,10])
	[0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
	>>> computeTicPositions([4,10])
	[4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0]
	>>> computeTicPositions([5,10])
	[5.0, 5.5, 6.0, 6.5, 7.0, 7.5, 8.0, 8.5, 9.0, 9.5, 10.0]
	>>> computeTicPositions([-5,10])
	[-5.0, -2.5, 0.0, 2.5, 5.0, 7.5, 10.0]
	>>> computeTicPositions([-4.5,10])
	[-2.5, 0.0, 2.5, 5.0, 7.5, 10.0]
	>>> map(str,computeTicPositions([0.25,1]))
	['0.3', '0.4', '0.5', '0.6', '0.7', '0.8', '0.9', '1.0']
	"""
	rangeDimension = math.floor(math.log10((valRange[1]-valRange[0])/10.))
	curTicDist = 10.**rangeDimension
	while (valRange[1]-valRange[0])/curTicDist>maxTics:
		if abs(curTicDist-10**math.floor(math.log10(curTicDist)))<1e-10:
			curTicDist *= 2.5
		else:
			curTicDist *= 2
	ticBase = math.floor(valRange[0]/curTicDist)*curTicDist
	return [s for s in
			[ticBase+curTicDist*i for i in range(100)]
		if valRange[0]-1e-10<=s<=valRange[1]+1e-10]


class PlottingCanvas(Tkinter.Canvas):
	"""This class encapsulates a canvas to plot functions.

	Whne construction a canvas, you have to specify the range
	in xRange and yRange, which are supposed to be sequences of
	length 2).  PlottingCanvas then takes care to draw a sufficiently
	good coordinate system.  You can then add functions to plot
	(these just need to be callable objects taking one float and
	returning one float).  You could even make them mutable and call
	redrawFunctions to update the display.

	Right now, you cannot remove single functions -- either all of
	them (clearFunctions()) or none.

	Kids: This ain't an interesting class.  It's not even beautiful.
	"""
	def __init__(self, master, xRange, yRange):
		Tkinter.Canvas.__init__(self, master)
		self.xRange = xRange
		self.yRange = yRange
		self.borderWidth = 20
		self.abscissaSpacing = 2
		self.nodeAbscissas = []
		self.functions = []
		self.functionItems = []
		self.bind("<Configure>", self._redraw)
		self.bind("<Motion>", self._reportPosition)
		self.config(cursor="crosshair")

	def _clearCanvas(self):
		self.delete(Tkinter.ALL)

	def _computeScalingAttributes(self, plotXSize, plotYSize, scaleX, scaleY):
		self.logToPhys = lambda x, y: (
			(x-self.xRange[0])/scaleX+self.borderWidth,
			plotYSize-((y-self.yRange[0])/scaleY-self.borderWidth))
		self.physToLog = lambda x, y: (
			(x-self.borderWidth)*scaleX+self.xRange[0],
			(plotYSize-y+self.borderWidth)*scaleY+self.yRange[0])
		self.isInDrawableArea = lambda x, y: (
			self.borderWidth<x<self.borderWidth+plotXSize and
			self.borderWidth<y<self.borderWidth+plotYSize)
		self.nodeAbscissas = [physX*scaleX
			for physX in range(0, plotXSize, self.abscissaSpacing)]

	def _getScalingParameters(self):
		xSize, ySize = self.winfo_width(), self.winfo_height()
		self.posDisplay = self.create_text(xSize, 0, anchor=Tkinter.NE)
		plotXSize = xSize-2*self.borderWidth
		plotYSize = ySize-2*self.borderWidth
		scaleX = (self.xRange[1]-self.xRange[0])/float(plotXSize)
		scaleY = (self.yRange[1]-self.yRange[0])/float(plotYSize)
		return plotXSize, plotYSize, scaleX, scaleY

	def _redraw(self, ev=None):
		self._clearCanvas()
		self._computeScalingAttributes(
			*self._getScalingParameters())
		self._drawAxes()
		self.redrawFunctions()

	def _drawLogicalLine(self, *points):
		coords = []
		for point in points:
			coords.extend(self.logToPhys(*point))
		return self.create_line(*coords)

	def _addLogicalText(self, coords, text, **kwargs):
		self.create_text(self.logToPhys(*coords), text=text, **kwargs)

	def _drawAxes(self):
		self._drawLogicalLine((self.xRange[0], self.yRange[0]),
			(self.xRange[0], self.yRange[1]))
		self._drawLogicalLine((self.xRange[0], self.yRange[0]),
			(self.xRange[1], self.yRange[0]))
		self._drawLogicalLine((self.xRange[0], self.yRange[1]),
			(self.xRange[1], self.yRange[1]))
		self._drawLogicalLine((self.xRange[1], self.yRange[0]),
			(self.xRange[1], self.yRange[1]))
		xTicLength = (self.yRange[1]-self.yRange[0])/50.
		for xTicPos in computeTicPositions(self.xRange):
			self._drawLogicalLine((xTicPos, self.yRange[0]),
				(xTicPos, self.yRange[0]+xTicLength))
			self._addLogicalText((xTicPos, self.yRange[0]), "%g"%xTicPos,
				anchor=Tkinter.N)
		yTicLength = (self.xRange[1]-self.xRange[0])/50.
		for yTicPos in computeTicPositions(self.yRange):
			self._drawLogicalLine((self.xRange[0], yTicPos),
				(self.xRange[0]+yTicLength, yTicPos))
			self._addLogicalText((self.xRange[0], yTicPos), "%g"%yTicPos,
				anchor=Tkinter.E)

	def _reportPosition(self, ev):
		if self.isInDrawableArea(ev.x, ev.y):
			text = "%1.3g/%1.3g"%self.physToLog(ev.x, ev.y)
		else:
			text = ""
		self.itemconfigure(self.posDisplay, text=text)

	def redrawFunctions(self):
		if not self.nodeAbscissas:
			return
		for id in self.functionItems:
			self.delete(id)
		funPoints = [[] for i in range(len(self.functions))]
		for x in self.nodeAbscissas:
			for funIndex, function in enumerate(self.functions):
				funPoints[funIndex].extend(self.logToPhys(
					x, function(x)))
		for coords in funPoints:
			self.functionItems.append(self.create_line(*coords))

	def addFunction(self, function):
		self.functions.append(function)
		self.redrawFunctions()
	
	def clearFunctions(self):
		self.functions = []
		self.redrawFunctions()
	

class TestController(Tkinter.Frame):
	"""This class encapsulates a graphic controller for the parameters of
	the test model.

	Kids: Not interesting, either.
	"""
	defaults = {
		"numExper": "20",
		"critVal": "19",
	}

	def __init__(self, master, **kwargs):
		Tkinter.Frame.__init__(self, master, **kwargs)
		self._buildWidgets()
	
	def _buildWidgets(self):
		self.entries = {}
		Tkinter.Label(self, text="n").pack(side=Tkinter.LEFT)
		self.entries["numExper"] = Tkinter.Entry(self, width=3)
		self.entries["numExper"].pack(side=Tkinter.LEFT)
		self.entries["numExper"].bind("<Return>", self._updateModel)
		self.entries["numExper"].bind("<FocusOut>", self._updateModel)
		Tkinter.Label(self, text="t").pack(side=Tkinter.LEFT)
		self.entries["critVal"] = Tkinter.Entry(self, width=3)
		self.entries["critVal"].pack(side=Tkinter.LEFT)
		self.entries["critVal"].bind("<Return>", self._updateModel)
		self.entries["critVal"].bind("<FocusOut>", self._updateModel)
		Tkinter.Button(self, text="Set", command=self._updateModel
			).pack(side=Tkinter.LEFT)

	def _readAnInt(self, entryName):
		try:
			return int(self.entries[entryName].get())
		except:
			self.entries[entryName].delete(0, Tkinter.END)
			self.entries[entryName].insert(0, self.defaults[entryName])
			return int(self.defaults[entryName])

	def _readARange(self, entryName):
		def parseRange(rawRange):
			startVal, endVal = map(int, rawRange.split("-"))
			return range(startVal, endVal+1)
		try:
			return parseRange(self.entries[entryName].get())
		except:
			self.entries[entryName].delete(0, Tkinter.END)
			self.entries[entryName].insert(0, self.defaults[entryName])
			return parseRange(self.defaults[entryName])

	def _updateModel(self, ev=None):
		self.postChanges(self.getValDict())

	def getValDict(self):
		valDict = {}
		valDict["numExper"] = self._readAnInt("numExper")
		valDict["rejRange"] = range(self._readAnInt("critVal"),
			valDict["numExper"]+1)
		return valDict

	def postChanges(self, valDict):
		self.master.postChanges(valDict)


class TestExplorator(Tkinter.Tk):
	"""This class encapsulates the aggregator for model, view and controller
	of our test.

	Kids: another class to ignore.
	"""
	def __init__(self, **kwargs):
		Tkinter.Tk.__init__(self, **kwargs)
		self.title("Binomial Test Toy")
		self._buildWidgets()
		self.model = ExactBinomialTest(**self.controller.getValDict())
		# draw function once we're mapped
		self.bind("<Map>", lambda ev: self.postChanges(
			self.controller.getValDict()))

	def _buildWidgets(self):
		self.view = PlottingCanvas(self, [0,1], [0,1])
		self.controller = TestController(self)
		self.view.grid(row=0, column=0, columnspan=2, 
			sticky=Tkinter.SW+Tkinter.NE)

		self.controller.grid(row=1, column=0, sticky=Tkinter.SW)
		Tkinter.Button(self, text="Quit", command=self.quit).grid(
			row=1, column=1, sticky=Tkinter.SE)
		self.columnconfigure(0, weight=1)
		self.rowconfigure(0, weight=1)

	def postChanges(self, valDict):
		self.model.updateValues(valDict)
		self.view.clearFunctions()
		self.view.addFunction(self.model.getPowerFunction())


class InterpolatingFunction:
	"""This class implements a function defined by linear interpolation between
	a couple of nodes.

	We don't do extrapolation (and punish requests to do so with IndexErrors),
	you have keep your abscissas between the end points of your node list.

	It may be a good exercise to write a class doing cubic interpolation.
	>>> f = InterpolatingFunction([(0,0), (1,1), (2,0.5)])
	>>> f(0)
	0
	>>> print f(0.5)
	0.5
	>>> print f(1.2)
	0.9
	>>> f(2.1)
	Traceback (most recent call last):
		...
	IndexError: list index out of range
	>>> f(-0.1)
	Traceback (most recent call last):
		...
	IndexError: list index out of range
	"""
	def __init__(self, nodes):
		self.nodes = nodes[:]
		self.nodes.sort()
	
	def __call__(self, abscissa):
		index = bisect.bisect_left(self.nodes, (abscissa, 0))
		if abs(self.nodes[index][0]-abscissa)<1e-10:
			return self.nodes[index][1]
		if index==0:  # -1 is a valid index, unfortunately
			raise IndexError("list index out of range")
		numerator = float(
			self.nodes[index][0]-self.nodes[index-1][0])
		if numerator==0:
			return self.nodes[index][1]
		slope = (self.nodes[index][1]-self.nodes[index-1][1]
			)/numerator
		return self.nodes[index][1]+slope*(abscissa-
			self.nodes[index][0])


class ExactBinomialTest:
	"""This class encapsulates our model, i.e. the test.

	You construct a test giving the number of individual trial for each test
	and a set of results the hypothesis should be rejected for.

	The class then pre-computes a number of binomials (right now, we have
	fixed the number of discretisation steps to 30).  These can then be used
	to get a power function.
	"""
	def __init__(self, numExper, rejRange):
		self.rejRange = rejRange
		self.numExper = numExper
		self._setBinomials()
	
	def _setBinomials(self, discSteps=30):
		"""pre-computes binomials for p=0...1.

		discSteps is the number of binomials to compute.  This number will
		be used as the number of nodes for the computation of the power
		function.
		"""
		self.binomials = {}
		for p in [i/float(discSteps-1) 
				for i in range(discSteps)]:
			self.binomials[p] = [binomialDist(p, i, self.numExper)
				for i in range(self.numExper+1)]

	def updateValues(self, valDict):
		"""updates the number of trials per test and the rejection range.

		valDict may give values for any combination of the parameters numExper
		and rejRange.  updateValues takes care that any cached values will
		be updated.
		"""
		if self.numExper!=valDict.get("numExper", self.numExper):
			self.numExper = valDict["numExper"]
			self._setBinomials()
		self.rejRange = valDict.get("rejRange", self.rejRange)

	def getPowerFunction(self):
		"""returns a callable objects that returns the power function on the
		domain [0..1].
		"""
		valueList = []
		for theta, pTheta in self.binomials.iteritems():
			valueList.append((theta, sum([pTheta[rInd]
				for rInd in self.rejRange])))
		valueList.sort()
		return InterpolatingFunction(valueList)


def _test():
	import doctest, simpletest
	doctest.testmod(simpletest)

if __name__=="__main__":
	_test()
	te = TestExplorator()
	te.mainloop()
